from ..utils.reader import Reader
from ..utils.writer import Writer


class Decoder(Reader):
    def __init__(self, initial_bytes: bytes, header: dict):
        super().__init__(initial_bytes)

        # Variables
        self.readed = {}
        vertices = []
        bind_matrix = []
        joints = []
        vertex_weights = []
        weights = []
        vcount = []
        materials = []
        # Variables

        name = self.readString()
        group = self.readString()
        if header['version'] < 2:
            matrix = []
            for x in range(4):
                temp_list = []
                for x1 in range(4):
                    temp_list.append(self.readFloat())
                matrix.append(temp_list)
        vertex_count = self.readUByte()
        for x in range(vertex_count):
            vertex = []
            vertex_type = self.readString()
            vertex_index = self.readUByte()
            vertex_shorts = self.readUShort()
            vertex_scale = self.readFloat()
            vertex_count = self.readUInt32()

            if vertex_type == 'VERTEX':
                vertex_type = 'POSITION'

            for x1 in range(vertex_count):
                coordinates_massive = []
                for x2 in range(vertex_shorts):
                    coordinate = self.readShort()
                    coordinates_massive.append(coordinate / 32512)
                if vertex_type == 'TEXCOORD':
                    coordinates_massive[1::2] = [1-x for x in coordinates_massive[1::2]]
                vertex.append(coordinates_massive)

            vertices.append({'type': vertex_type, 'index': vertex_index, 'scale': vertex_scale, 'vertex': vertex})
        have_bind_matrix = self.readBool()
        if have_bind_matrix:
            for x in range(16):
                bind_matrix.append(self.readFloat())
        joint_counts = self.readUByte()
        for x in range(joint_counts):
            joint_matrix = []
            joint_name = self.readString()
            for x1 in range(16):
                joint_matrix.append(self.readFloat())
            joints.append({'name': joint_name, 'matrix': joint_matrix})
        vertex_weights_count = self.readUInt32()
        for x in range(vertex_weights_count):
            vcount.append(0)
            joint_a = self.readUByte()
            joint_b = self.readUByte()
            joint_c = self.readUByte()
            joint_d = self.readUByte()
            weight_a = self.readUShort()
            weight_b = self.readUShort()
            weight_c = self.readUShort()
            weight_d = self.readUShort()
            temp_list = [[joint_a, weight_a], [joint_b, weight_b], [joint_c, weight_c], [joint_d, weight_d]]
            for pair in temp_list:
                if pair[1] != 0:
                    vcount[x] += 1
                    vertex_weights.append(pair[0])
                    if pair[1]/65535 not in weights:
                        weights.append(pair[1]/65535)
                    vertex_weights.append(weights.index(pair[1]/65535))
        materials_count = self.readUByte()
        for x in range(materials_count):
            polygons = []
            material_name = self.readString()
            self.readString()
            polygons_count = self.readUShort()
            inputs_count = self.readUByte()
            vertex_id_length = self.readUByte()
            for x1 in range(polygons_count):
                temp_list = []
                for x2 in range(3):
                    second_temp_list = []
                    for x3 in range(inputs_count):
                        second_temp_list.append(self.readInteger(vertex_id_length))
                    temp_list.append(second_temp_list)
                polygons.append(temp_list)
            materials.append({'name': material_name, 'polygons': polygons})

        self.readed['name'] = name
        self.readed['group'] = group
        self.readed['vertices'] = vertices
        self.readed['have_bind_matrix'] = have_bind_matrix
        if have_bind_matrix:
            self.readed['bind_matrix'] = bind_matrix
            self.readed['joints'] = joints
            self.readed['weights'] = {}
            self.readed['weights']['vcount'] = vcount
            self.readed['weights']['weights'] = weights
            self.readed['weights']['vertex_weights'] = vertex_weights
        self.readed['materials'] = materials


class Encoder(Writer):
    def __init__(self, data: dict):
        super().__init__()
        self.name = 'GEOM'
        self.data = data

        self.encode()

        self.length = len(self.buffer)
        
    def encode(self):
        self.writeString(self.data['name'])
        self.writeString(self.data['group'])

        self.encode_vertices(self.data['vertices'])

        self.writeBool(self.data['have_bind_matrix'])
        if self.data['have_bind_matrix']:
            for x in self.data['bind_matrix']:
                self.writeFloat(x)

        self.encode_joints()

        self.encode_weight()

        self.encode_triangles()

    def encode_vertices(self, vertices: dict):
        self.writeUByte(len(vertices))
        for vertex in vertices:
            self.writeString(vertex['type'])
            self.writeUByte(vertex['index'])
            self.writeUShort(len(vertex['vertex'][0]))
            self.writeFloat(vertex['scale'])
            self.writeUInt32(len(vertex['vertex']))
            for coordinates_massive in vertex['vertex']:
                if vertex['type'] == 'TEXCOORD':
                    coordinates_massive[1::2] = [1 - x for x in coordinates_massive[1::2]]
                for coordinate in coordinates_massive:
                    coordinate /= vertex['scale']
                    coordinate *= 32512
                    self.writeShort(round(coordinate))

    def encode_joints(self):
        if self.data['have_bind_matrix']:
            self.writeUByte(len(self.data['joints']))

            for joint in self.data['joints']:
                self.writeString(joint['name'])
                for x in joint['matrix']:
                    self.writeFloat(x)
        else:
            self.writeUByte(0)

    def encode_weight(self):
        if self.data['have_bind_matrix']:
            self.writeUInt32(len(self.data['weights']['vcount']))
            past_index = 0
            for vcount in self.data['weights']['vcount']:
                temp_list = []
                for x in range(vcount):
                    vertex_weights_index = x * 2 + past_index * 2
                    joint_id = self.data['weights']['vertex_weights'][vertex_weights_index]
                    weight_id = self.data['weights']['vertex_weights'][vertex_weights_index + 1]

                    weight = self.data['weights']['weights'][weight_id]

                    if weight > 1:
                        weight = 1
                    elif weight < 1:
                        weight = 0

                    weight = int(weight * 65535)
                    temp_list.append([joint_id, weight])
                past_index += vcount
                while len(temp_list) < 4:
                    temp_list.append([0, 0])
                for x in temp_list:
                    self.writeUByte(x[0])
                for x in temp_list:
                    self.writeUShort(x[1])
        else:
            self.writeUInt32(0)

    def encode_triangles(self):
        self.writeUByte(len(self.data['materials']))
        for material in self.data['materials']:
            self.writeString(material['name'])
            self.writeString('')
            self.writeUShort(len(material['polygons']))

            # Calculate settings
            inputs_count = len(material['polygons'][0][0])

            maximal_value = 0
            for points in material['polygons']:
                for point in points:
                    for vertex in point:
                        if vertex > maximal_value:
                            maximal_value = vertex

            short_length = 1 if maximal_value <= 127 else 2

            # Write Settings
            self.writeUByte(inputs_count)
            self.writeUByte(short_length)

            # Write Polygons
            for points in material['polygons']:
                for point in points:
                    for vertex in point:
                        self.writeInteger(vertex, short_length)



from ..utils.reader import Reader
from ..utils.writer import Writer


class Decoder(Reader):
    def __init__(self, initial_bytes: bytes, header: dict):
        super().__init__(initial_bytes)

        self.readed = {}

        name = self.readString()
        shader = self.readString()
        self.readUByte()
        self.readed['name'] = name
        self.readed['shader'] = shader

        self.readed['effect'] = {}
        use_ambient_tex = self.readBool()
        if use_ambient_tex:
            ambient_tex = self.readString()
            self.readed['effect']['ambient'] = ambient_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            ambient_color = (r, g, b, a)
            self.readed['effect']['ambient'] = ambient_color

        use_diffuse_tex = self.readBool()
        if use_diffuse_tex:
            diffuse_tex = self.readString()
            self.readed['effect']['diffuse'] = diffuse_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            diffuse_color = (r, g, b, a)
            self.readed['effect']['diffuse'] = diffuse_color

        use_specular_tex = self.readBool()
        if use_specular_tex:
            specular_tex = self.readString()
            self.readed['effect']['specular'] = specular_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            specular_color = (r, g, b, a)
            self.readed['effect']['specular'] = specular_color

        texture_one = self.readString()
        texture_two = self.readString()

        use_colorize_tex = self.readBool()
        if use_colorize_tex:
            colorize_tex = self.readString()
            self.readed['effect']['colorize'] = colorize_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            colorize_color = (r, g, b, a)
            self.readed['effect']['colorize'] = colorize_color

        use_emission_tex = self.readBool()
        if use_emission_tex:
            emission_tex = self.readString()
            self.readed['effect']['emission'] = emission_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            emission_color = (r, g, b, a)
            self.readed['effect']['emission'] = emission_color

        self.readString()
        self.readFloat()
        self.readFloat()

        lightmap_diffuse_tex = self.readString()
        lightmap_specular_tex = self.readString()
        self.readed['effect']['lightmaps'] = {
            'diffuse': lightmap_diffuse_tex,
            'specular': lightmap_specular_tex
        }

        a = self.readUByte()
        r = self.readUByte()
        g = self.readUByte()
        b = self.readUByte()
        tint = (r, g, b, a)
        self.readed['effect']['tint'] = tint


class Encoder(Writer):
    def __init__(self, data: dict):
        super().__init__()
        self.name = 'MATE'
        self.data = data

        self.encode()

        self.length = len(self.buffer)

    def encode(self):
        pass



from .matrix3x3 import Matrix3x3


def matrix2tuple(matrix_list: list):
    new_matrix = [tuple(row) for row in matrix_list]
    new_matrix = tuple(new_matrix)

    return new_matrix


class Matrix4x4(object):
    def __init__(self, matrix_list: tuple = None):
        self.translation_matrix = None
        self.rotation_matrix = None
        self.scale_matrix = None

        if matrix_list is not None:
            self.matrix = matrix_list
        else:
            self.matrix = (
                (1, 0, 0, 0),  # x
                (0, 1, 0, 0),  # y
                (0, 0, 1, 0),  # z
                (0, 0, 0, 1)
            )

    def __matmul__(self, other):
        if len(self.matrix) != len(other.matrix[0]) or len(self.matrix[0]) != len(other.matrix):
            raise TypeError('Матрицы не могут быть перемножены')
        else:
            multiplied_matrix = []

            for row in self.matrix:
                matrix_row = []
                for column in range(4):
                    s = 0
                    for i in range(4):
                        s += row[i] * other.matrix[i][column]
                    matrix_row.append(s)
                multiplied_matrix.append(matrix_row)

            return Matrix4x4(matrix2tuple(multiplied_matrix))

    def __mul__(self, other: int or float):
        new_matrix = []
        for row in range(len(self.matrix)):
            new_row = []
            for column in range(len(self.matrix[0])):
                new_row.append(self.matrix[row][column]*other)
            new_matrix.append(new_row)

        return Matrix4x4(matrix2tuple(new_matrix))

    def __str__(self):
        return str(self.matrix)

    def determinant(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]
        value1_3 = self.matrix[0][2]
        value1_4 = self.matrix[0][3]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]
        value2_3 = self.matrix[1][2]
        value2_4 = self.matrix[1][3]

        value3_1 = self.matrix[2][0]
        value3_2 = self.matrix[2][1]
        value3_3 = self.matrix[2][2]
        value3_4 = self.matrix[2][3]

        value4_1 = self.matrix[3][0]
        value4_2 = self.matrix[3][1]
        value4_3 = self.matrix[3][2]
        value4_4 = self.matrix[3][3]

        matrix1 = Matrix3x3((
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix2 = Matrix3x3((
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix3 = Matrix3x3((
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix4 = Matrix3x3((
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        det = value1_1 * matrix1.determinant()
        det -= value1_2 * matrix2.determinant()
        det += value1_3 * matrix3.determinant()
        det -= value1_4 * matrix4.determinant()

        return det

    def transpose(self):
        height = len(self.matrix)
        width = len(self.matrix[0])

        self.matrix = [[self.matrix[row][col] for row in range(height)] for col in range(width)]
        self.matrix = [tuple(row) for row in self.matrix]
        self.matrix = tuple(self.matrix)

    def cofactor(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]
        value1_3 = self.matrix[0][2]
        value1_4 = self.matrix[0][3]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]
        value2_3 = self.matrix[1][2]
        value2_4 = self.matrix[1][3]

        value3_1 = self.matrix[2][0]
        value3_2 = self.matrix[2][1]
        value3_3 = self.matrix[2][2]
        value3_4 = self.matrix[2][3]

        value4_1 = self.matrix[3][0]
        value4_2 = self.matrix[3][1]
        value4_3 = self.matrix[3][2]
        value4_4 = self.matrix[3][3]

        matrix1 = Matrix3x3((
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix2 = Matrix3x3((
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix3 = Matrix3x3((
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix4 = Matrix3x3((
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix5 = Matrix3x3((
            (value1_2, value1_3, value1_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix6 = Matrix3x3((
            (value1_1, value1_3, value1_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix7 = Matrix3x3((
            (value1_1, value1_2, value1_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix8 = Matrix3x3((
            (value1_1, value1_2, value1_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix9 = Matrix3x3((
            (value1_2, value1_3, value1_4),
            (value2_2, value2_3, value2_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix10 = Matrix3x3((
            (value1_1, value1_3, value1_4),
            (value2_1, value2_3, value2_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix11 = Matrix3x3((
            (value1_1, value1_2, value1_4),
            (value2_1, value2_2, value2_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix12 = Matrix3x3((
            (value1_1, value1_2, value1_3),
            (value2_1, value2_2, value2_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix13 = Matrix3x3((
            (value1_2, value1_3, value1_4),
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4)
        ))

        matrix14 = Matrix3x3((
            (value1_1, value1_3, value1_4),
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4)
        ))

        matrix15 = Matrix3x3((
            (value1_1, value1_2, value1_4),
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4)
        ))

        matrix16 = Matrix3x3((
            (value1_1, value1_2, value1_3),
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3)
        ))

        self.matrix = (
            (matrix1.determinant(), matrix2.determinant(), matrix3.determinant(), matrix4.determinant()),
            (matrix5.determinant(), matrix6.determinant(), matrix7.determinant(), matrix8.determinant()),
            (matrix9.determinant(), matrix10.determinant(), matrix11.determinant(), matrix12.determinant()),
            (matrix13.determinant(), matrix14.determinant(), matrix15.determinant(), matrix16.determinant())
        )

        cofactor_matrix = []
        for row in range(len(self.matrix)):
            new_row = []
            for column in range(len(self.matrix[0])):
                new_row.append(self.matrix[row][column] * (-1) ** (row + column))
            cofactor_matrix.append(new_row)
        self.matrix = cofactor_matrix

    def inverse(self):
        det = self.determinant()
        if det != 0:
            self.transpose()
            self.cofactor()

            self.matrix = self.__mul__(1 / det).matrix

        return self

    def put_rotation(self, xyz: tuple, w: float):
        x, y, z = xyz

        rotation_matrix = (
            (1-2*y**2-2*z**2, 2*x*y-2*z*w, 2*x*z+2*y*w, 0),  # x
            (2*x*y+2*z*w, 1-2*x**2-2*z**2, 2*y*z-2*x*w, 0),  # y
            (2*x*z-2*y*w, 2*y*z+2*x*w, 1-2*x**2-2*y**2, 0),  # z
            (0, 0, 0, 1)
        )

        self.rotation_matrix = Matrix4x4(rotation_matrix)

    def put_position(self, xyz: tuple):
        x, y, z = xyz

        translation_matrix = (
            (1, 0, 0, x),  # x
            (0, 1, 0, y),  # y
            (0, 0, 1, z),  # z
            (0, 0, 0, 1)
        )

        self.translation_matrix = Matrix4x4(translation_matrix)

    def put_scale(self, xyz: tuple):
        x, y, z = xyz

        scale_matrix = Matrix4x4((
            (x, 0, 0, 0),
            (0, y, 0, 0),
            (0, 0, z, 0),
            (0, 0, 0, 1)
        ))

        self.scale_matrix = scale_matrix

    def get_rotation(self):
        pass

    def get_position(self):
        xyz = (self.matrix[0][3], self.matrix[1][3], self.matrix[2][3])

        self.put_position(xyz)

        return xyz

    def get_scale(self):
        xyz = (1, 1, 1)

        self.put_scale(xyz)

        return xyz



from ..utils.reader import Reader
from ..utils.writer import Writer


class Decoder(Reader):
    def __init__(self, initial_bytes: bytes, header: dict):
        super().__init__(initial_bytes)

        self.readed = {}


class Encoder(Writer):
    def __init__(self, data: dict):
        super().__init__()
        self.name = 'WEND'
        self.data = data

        self.encode()

        self.length = len(self.buffer)

    def encode(self):
        pass



import setuptools

with open('README.md') as fh:
    long_description = fh.read()
    fh.close()

setuptools.setup(
    name='3d-converter',
    version='0.4.9',
    author='Vorono4ka',
    author_email='crowo4ka@gmail.com',
    description='Python 3D Models Converter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/vorono4ka/3d-converter',
    license='GPLv3',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
